import sys
from pathlib import Path

import click

from workspace.cli import theme, utils
from workspace.core.adapter import get_adapters
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
)
@click.option("--type", type=str, help="The type of project to add.")
@click.option(
    "--name", type=str, default=None, help="Name of the project. If not provided, the directory name will be used."
)
def add(path: Path, type: str = None, name: str = None):
    """Track an existing project at PATH in the workspace."""
    workspace = Workspace.from_path()

    name = name or path.name
    if name in workspace.projects:
        theme.echo(
            f"""<e>Project <b>{name}</b> already exists.</e>

Specify a different name with:

    <a>workspace add{f" --type {type}" if type else ""} <b>--name NAME</b> {path}
"""
        )
        sys.exit(1)

    existing = workspace.get_project_by_path(path)
    if existing:
        theme.echo(f"<e>Path <b>{path}</b> already tracked as project <b>{existing.name}</b>.</e>")
        sys.exit(1)

    type = type or utils.detect_type(workspace, path)
    if not type:
        valid_type_list = "\n".join([f"  - <b>{name}</b>" for name in get_adapters()])
        theme.echo(
            f"""<e>Could not detect type of project at path <b>{path}</b>.</e>

Please specify type as follows:

    <a>workspace add --type <b>TYPE</b> {path}</a>

Available types are:
{valid_type_list}
"""
        )
        sys.exit(1)

    project = workspace.set_project(
        name=name,
        path=str(Path.cwd().relative_to(workspace.path) / path),
        type=type,
    )
    project.adapter.validate()
    workspace.flush()
    theme.echo(f"Added new project <a>{project.name}</a> at <a>{path}</a>.")



import shutil
import sys
from typing import Tuple

import click

from workspace.cli import callbacks, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option("--delete/--no-delete", "-d/ ", type=bool, default=False, help="Delete the workspace content too.")
def remove(specifiers: Tuple[str, ...], delete: bool = False):
    """Remove project(s) from the workspace.

    By default the project files will not be deleted, just untracked. Add the --delete flag to remove the directory.
    """
    workspace = Workspace.from_path()

    target_set = resolve_specifiers(workspace, specifiers)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    for name in target_set:
        resolved_path = workspace.projects[name].resolved_path
        del workspace.projects[name]
        workspace.flush()
        if delete:
            shutil.rmtree(resolved_path)

    names = "\n".join(f"  - <b>{name}</b>" for name in target_set)
    if delete:
        theme.echo(
            f"""<w>Deleted projects:</w>
{names}
"""
        )
    else:
        theme.echo(
            f"""Stopped tracking projects:
{names}
"""
        )



import sys
from collections import defaultdict
from functools import lru_cache
from typing import DefaultDict, Dict, Set, Tuple

import click

from workspace.cli import callbacks, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option("--transitive/--no-transitive", type=bool, default=True, help="Only show direct dependencies.")
@click.option(
    "--output",
    "-o",
    type=click.Choice(("lines", "csv")),
    help="Select the output format.",
    default="lines",
)
@click.option(
    "--dev/--no-dev",
    "-d/ ",
    type=bool,
    help="Include development dependencies.",
    default=False,
)
def dependencies(specifiers: Tuple[str, ...], transitive: bool = True, output: str = "lines", dev: bool = False):
    """Get the set of all projects which are dependencies of the specified projects."""
    workspace = Workspace.from_path()

    target_set = resolve_specifiers(workspace, specifiers)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    # Get a full map of dependencies:
    dependency_map = _get_dependency_map(workspace, transitive=transitive, include_dev=dev)

    # Extract the relevant dependencies:
    dependencies_set = {dependency for target in target_set for dependency in dependency_map[target]}

    # Sort them from least dependencies, to most dependencies:
    sorted_dependencies = [
        name
        for name, _ in sorted(dependency_map.items(), key=lambda item: len(item[1]))
        if name in dependencies_set or name in target_set
    ]

    if output == "csv":
        theme.echo(",".join(sorted_dependencies), err=False)
        sys.exit(0)

    for dependency in sorted_dependencies:
        theme.echo(dependency, err=False)
    sys.exit(0)


def _get_dependency_map(
    workspace: Workspace,
    transitive: bool = True,
    include_dev: bool = False,
) -> Dict[str, Set[str]]:
    """Get a mapping of projects to their dependencies."""

    direct_map: DefaultDict[str, Set[str]] = defaultdict(set)
    for name, project in workspace.projects.items():
        direct_map[name] |= project.adapter.dependencies(include_dev=include_dev)
    if not transitive:
        return dict(direct_map)

    @lru_cache(maxsize=None)
    def get_dependencies(name: str):
        dependencies = set(direct_map[name])
        for dependency in set(dependencies):
            dependencies |= get_dependencies(dependency)
        return dependencies

    return {name: get_dependencies(name) for name in workspace.projects}



import importlib

import click

from workspace.cli import theme
from workspace.cli.exceptions import WorkspaceCLIError
from workspace.core.adapter import Adapter
from workspace.core.models import Workspace


@click.group()
def plugin():
    """Commands for managing workspace plugins."""
    pass


@plugin.command()
@click.argument("module_path", type=str)
def add(module_path: str):
    """Add a plugin by its Python module path."""
    workspace = Workspace.from_path()

    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError:
        raise WorkspaceCLIError(
            f"""<e>Could not import plugin <b>{module_path}</b>.</e>

The plugin must be available on your <a>PYTHONPATH</a>.
"""
        )
    workspace.plugins = workspace.plugins or []
    if module_path in workspace.plugins:
        raise WorkspaceCLIError(f"<e>Plugin <b>{module_path}</b> already installed.</e>")
    workspace.plugins.append(module_path)
    workspace.flush()
    theme.echo(f"Added plugin <s>{module_path}</s>.")

    new_types = "\n".join(
        [
            f"  - <a>{cls.name}</a>"
            for cls in vars(module).values()
            if isinstance(cls, type) and issubclass(cls, Adapter) and not cls is Adapter
        ]
    )
    if new_types:
        theme.echo(
            f"""
<h>The following new project types are now available:</h>
{new_types}
"""
        )


@plugin.command()
@click.argument("module_path", type=str)
def remove(module_path: str):
    """Remove a plugin by its Python module path."""
    workspace = Workspace.from_path()

    workspace.plugins = workspace.plugins or []
    if module_path not in workspace.plugins:
        raise WorkspaceCLIError(f"<e>Plugin <b>{module_path}</b> not installed.</e>")
    workspace.plugins.remove(module_path)
    workspace.flush()
    theme.echo(f"Removed plugin <s>{module_path}</s>.")


@plugin.command("list")
def list_():
    workspace = Workspace.from_path()
    for plugin in workspace.plugins or []:
        theme.echo(plugin)



import time
from itertools import cycle
from typing import Dict, Iterable, List, NamedTuple, Tuple

from workspace.cli import theme
from workspace.core.models import Project


def run(project_commands: List[Tuple[Project, str]], *, parallel: bool) -> int:
    """Run each command in each project, in series or parallel."""
    if parallel:
        return _run_in_parallel(project_commands)
    return _run_in_series(project_commands)


def _run_in_series(project_commands: List[Tuple[Project, str]]) -> int:
    """Run each command in each project in series.

    Command output is not captured.
    """
    exit_codes: Dict[str, int] = {}
    for project, command in project_commands:
        theme.echo(f"\nRunning <a><b>{command}</b></a>  (<b>{project.name}</b>)\n")
        result = project.adapter.run(command)
        exit_codes[project.name] = result.returncode

    failed = {name for name, code in exit_codes.items() if code}

    theme.echo("")
    if failed:
        theme.echo("<e>Some projects failed</e>: " + ", ".join([f"<b>{name}</b>" for name in failed]))
    return _get_exit_code(exit_codes.values())


def _run_in_parallel(project_commands: List[Tuple[Project, str]]) -> int:
    """Run each command in each project in parallel.

    Command output is captured, and only displayed for projects which failed after all
    projects have finished.
    """
    commands = {command for _, command in project_commands}
    if len(commands) == 1:
        command_repr = commands.pop()
    else:
        command_repr = "[various commands]"
    theme.echo(f"\nRunning <a><b>{command_repr}</b></a>\n")
    running = {project.name: project.adapter.popen(command) for project, command in project_commands}
    complete: Dict[str, Result] = {}
    spinner = Spinner()
    timer = Timer()
    max_name_length = max([len(name) for name in running])
    while running:
        for name, popen in running.copy().items():
            exit_code = popen.poll()
            theme.echo(
                f"{next(spinner)} <a>Running ({next(timer)})</a>: " + ", ".join([f"<b>{name}</b>" for name in running]),
                nl=False,
                rewrite=True,
            )
            if exit_code is None:
                continue
            stdout, stderr = popen.communicate()
            del running[name]
            complete[name] = Result(exit_code=exit_code, stdout=stdout, stderr=stderr)
            if exit_code != 0:
                theme.echo(f"<e>✘ {name.ljust(max_name_length)} ({next(timer)})</e>", rewrite=True)
            else:
                theme.echo(f"<s>✔ {name.ljust(max_name_length)} ({next(timer)})</s>", rewrite=True)

    theme.echo("")
    for name, result in complete.items():
        if not result.success:
            theme.echo(f"<e><b>{name}</b> failed with exit code <b>{exit_code}</b></e>:")
            theme.echo(result.output)

    return _get_exit_code({result.exit_code for result in complete.values()})


def _get_exit_code(exit_codes: Iterable[int]) -> int:
    """Reduce a set of exit codes to the absolute value."""
    return sorted(exit_codes, key=lambda code: abs(code), reverse=True)[0]


def Spinner(chars: str = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏", fps: float = 10):
    """A spinner for showing progress."""
    frames = cycle(chars)
    current_frame = next(frames)
    last_update = time.monotonic()
    delay = 1 / fps
    while True:
        now = time.monotonic()
        if now - last_update > delay:
            last_update = now
            current_frame = next(frames)
        yield current_frame


def Timer():
    """A timer which shows total time elapsed on each iteration."""
    start = time.monotonic()
    while True:
        yield f"{time.monotonic() - start:.1f}s"


class Result(NamedTuple):
    """Result of a subprocess."""

    exit_code: int
    stdout: str
    stderr: str

    @property
    def output(self) -> str:
        return "\n".join([self.stdout, self.stderr])

    @property
    def success(self) -> bool:
        return self.exit_code == 0



import shutil
import sys
from pathlib import Path

import click

from workspace.cli import theme, utils
from workspace.cli.exceptions import WorkspaceCLIError
from workspace.core.adapter import get_adapter, get_adapters
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "path",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, path_type=Path),
)
@click.option("--type", type=str, help="The type of project to create.")
@click.option(
    "--name", type=str, default=None, help="Name of the project. If not provided, the directory name will be used."
)
@click.option(
    "--template",
    type=str,
    default=None,
    help="Specify a template to use. Must be discoverable on the template path. Optional if `type` is specified, and the type adapter provides a default template.",
)
def new(path: Path, type: str = None, name: str = None, template: str = None):
    """Create a new project in the workspace.

    Either specify a template to use, or specify the type to use the default template.
    """
    workspace = Workspace.from_path()

    existing = workspace.get_project_by_path(path)
    if existing:
        theme.echo(f"<e>Path <b>{path}</b> already tracked as project <b>{existing.name}</b>.</e>")
        sys.exit(1)

    name = name or path.name
    if path.resolve().exists():
        theme.echo(
            f"""<e>Path <b>{path}</b> already exists.</e>

Run the following to track an existing project:

    <a>workspace <b>add</b> {path}</a>
"""
        )
        sys.exit(1)

    if name in workspace.projects:
        theme.echo(
            f"""<e>Project <b>{name}</b> already exists.</e>

Specify a different name with:

    <a>workspace new{f" --type {type}" if type else ""}{f" --template {template}" if template else ""} <b>--name NAME</b> {path}
"""
        )
        sys.exit(1)

    if template and template not in workspace.templates:
        template_list = "\n".join([f"  - <a>{path.name}</a>" for path in workspace.templates])
        raise WorkspaceCLIError(
            f"""<e>Unknown template <b>{template}</b>.</e>.

Available templates are:
{template_list}

"""
        )

    try:
        _initialise_template(workspace, path=path.resolve(), type=type, template=template)
        type = type or utils.detect_type(workspace, path.resolve())
        if not type:
            valid_type_list = "\n".join([f"  - <b>{name}</b>" for name in get_adapters()])
            template_path = workspace.templates[template]  # type: ignore[index]
            raise WorkspaceCLIError(
                f"""
<e>Could not detect type of generated project.</e>

This is likely an issue with the template at <b>{template_path}</b>.

You can specify the intended type for more detailed information:

    <a>workspace add --template {template} --type <b>TYPE</b> {path}</a>

Available types are:
{valid_type_list}
"""
            )
        project = workspace.set_project(
            name=name,
            path=str(Path.cwd().relative_to(workspace.path) / path),
            type=type,
        )
        project.adapter.validate()
        workspace.flush()
    except Exception as exc:
        if path.exists():
            shutil.rmtree(path.resolve())
        raise exc
    else:
        theme.echo(f"Created new project <a>{name}</a> at <a>{path}</a>.")
        sys.exit(0)


def _initialise_template(workspace: Workspace, path: Path, type: str = None, template: str = None):
    """Initialise the project directory."""
    if template:
        workspace.templates.create(template, path)
    else:
        if not type:
            raise WorkspaceCLIError("<e>Must specify at least one of <b>--type</b> and <b>--template</b> options.</e>")
        adapter = get_adapter(type)
        try:
            adapter.new(path)
        except NotImplementedError:
            raise WorkspaceCLIError(
                f"""<e>Type <b>{type}</b> does not have a default template.</e>

Please provide one, e.g.

    <a>workspace new {path} --type {type} --template TEMPLATE</a>
"""
            )



from __future__ import annotations

import re
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Set

from workspace.core.exceptions import WorkspaceTemplateError

if TYPE_CHECKING:
    from workspace.core.models import Workspace  # pragma: no cover


@dataclass
class Templates:
    workspace: Workspace

    @property
    def template_paths(self) -> List[Path]:
        if not self.workspace.template_path:
            return []
        return [self.workspace.path / Path(path) for path in self.workspace.template_path]

    def __iter__(self) -> Iterator[Path]:
        seen: Set[str] = set()
        for template_path in self.template_paths:
            for path in template_path.glob("*/cookiecutter.json"):
                if path.parent.name in seen:
                    continue
                seen.add(path.parent.name)
                yield path.parent

    def __getitem__(self, template_name: str) -> Path:
        for path in self.template_paths:
            target = path / template_name / "cookiecutter.json"
            if target.exists() and target.is_file():
                return target.parent
        raise KeyError(template_name)

    def __contains__(self, template_name: str) -> bool:
        try:
            self[template_name]
            return True
        except KeyError:
            return False

    def create(self, template_name: str, path: Path):
        try:
            template_path = self[template_name]
        except KeyError:
            raise WorkspaceTemplateError(f"Template not found: {template_name}")
        try:
            from cookiecutter.main import cookiecutter
        except ImportError:
            raise WorkspaceTemplateError(
                "Cookiecutter is not installed - run 'pip install workspace-cli[cookiecutter]'."
            )
        cookiecutter(
            template=str(template_path),
            output_dir=path.parent,
            extra_context=_build_context(template_path, path),
            no_input=True,
        )


def _build_context(template_path: Path, target_directory: Path) -> str:
    """Build the cookiecutter context.

    We need to combine the target directory and prompt the user for all other fields. By
    default cookiecutter is all or nothing when it comes to user input, which is no good,
    because it would be easy to accidentally enter a value which conflicts with the already
    provided directory name.

    As such, we do our own prompt logic (heavily inspired by cookiecutter's).
    """
    from cookiecutter.generate import generate_context

    # Get the field name used for the template's directory
    directory_field = _extract_directory_field(template_path)

    # Prompt user to provide any other fields
    context = generate_context(
        context_file=template_path / "cookiecutter.json",
    )
    extra_context = _prompt_for_config(context, {directory_field: target_directory.name})

    # Return the rendered context
    return extra_context


def _prompt_for_config(context, initial_config: dict):  # pragma: no cover
    """Prompt user to enter a new config.

    This is copied from
    https://github.com/cookiecutter/cookiecutter/blob/b1f6427606b67362de233588dd7a37496195b031/cookiecutter/prompt.py#L180

    But doesn't prompt for already provided values.
    """
    from cookiecutter.environment import StrictEnvironment
    from cookiecutter.exceptions import UndefinedVariableInTemplate
    from cookiecutter.prompt import prompt_choice_for_config, read_user_dict, read_user_variable, render_variable
    from jinja2.exceptions import UndefinedError

    cookiecutter_dict = OrderedDict(initial_config)
    env = StrictEnvironment(context=context)

    # First pass: Handle simple and raw variables, plus choices.
    # These must be done first because the dictionaries keys and
    # values might refer to them.
    for key, raw in context["cookiecutter"].items():
        if key in cookiecutter_dict:
            continue
        if key.startswith("_") and not key.startswith("__"):
            cookiecutter_dict[key] = raw
            continue
        elif key.startswith("__"):
            cookiecutter_dict[key] = render_variable(env, raw, cookiecutter_dict)
            continue

        try:
            if isinstance(raw, list):
                # We are dealing with a choice variable
                val = prompt_choice_for_config(cookiecutter_dict, env, key, raw, False)
                cookiecutter_dict[key] = val
            elif not isinstance(raw, dict):
                # We are dealing with a regular variable
                val = render_variable(env, raw, cookiecutter_dict)

                val = read_user_variable(key, val)

                cookiecutter_dict[key] = val
        except UndefinedError as err:
            msg = "Unable to render variable '{}'".format(key)
            raise UndefinedVariableInTemplate(msg, err, context)

    # Second pass; handle the dictionaries.
    for key, raw in context["cookiecutter"].items():
        # Skip private type dicts not ot be rendered.
        if key in cookiecutter_dict:
            continue
        if key.startswith("_") and not key.startswith("__"):
            continue

        try:
            if isinstance(raw, dict):
                # We are dealing with a dict variable
                val = render_variable(env, raw, cookiecutter_dict)

                if not key.startswith("__"):
                    val = read_user_dict(key, val)

                cookiecutter_dict[key] = val
        except UndefinedError as err:
            msg = "Unable to render variable '{}'".format(key)
            raise UndefinedVariableInTemplate(msg, err, context)

    return cookiecutter_dict


def _extract_directory_field(template_path: Path) -> str:
    """Extract the name of the cookiecutter context field which controls the directory name."""
    from cookiecutter.find import find_template

    template_dir_name = Path(find_template(str(template_path))).name
    match = re.search(r"(?<=cookiecutter\.)[\w-]+", template_dir_name)
    if not match:
        raise WorkspaceTemplateError(f"Failed extracting context field from {template_dir_name!r}.")
    return match.group()



from __future__ import annotations

import shlex
import subprocess
from typing import TYPE_CHECKING, ClassVar, List, Optional, Set, Tuple

if TYPE_CHECKING:
    from pathlib import Path  # pragma: no cover

    from workspace.core.models import Project  # pragma: no cover


class Adapter:
    name: ClassVar[Optional[str]]
    command_prefix: ClassVar[List[str]]

    def __init_subclass__(cls, *, name: str = None, command_prefix: Tuple[str, ...] = None):
        cls.name = name
        cls.command_prefix = list(command_prefix or [])

    def __init__(self, project: Project):
        self._project = project

    def __repr__(self):
        return f"{type(self)}({self._project})"

    def run(self, command: str, capture_output: bool = False, check: bool = False) -> subprocess.CompletedProcess:
        """Run a command within the project environment."""
        with self.popen(command, capture_output=capture_output) as process:
            try:
                stdout, stderr = process.communicate()
            except:
                process.kill()
                raise
            returncode = process.returncode
        result: subprocess.CompletedProcess = subprocess.CompletedProcess(
            command,
            returncode=returncode,
            stdout=stdout,
            stderr=stderr,
        )
        if check:
            result.check_returncode()
        return result

    def popen(self, command: str, capture_output: bool = True) -> subprocess.Popen:
        """Open a subprocess and return it."""
        command, kwargs = self.run_args(command)
        if capture_output:
            kwargs.update(stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return subprocess.Popen(command, **kwargs)

    def run_args(self, command: str) -> Tuple[str, dict]:
        """Get modified command and kwargs that should be used when running inside the project.

        These arguments are passed directly to subprocess.Popen.
        """
        # We don't do shlex.join(prefix + shlex.split(command)) because this will escape e.g. |
        command = " ".join([shlex.join(self.command_prefix), command])
        return command, dict(cwd=self._project.resolved_path, shell=True)

    def dependencies(self, include_dev: bool = True) -> Set[str]:
        """Return the names of projects this project depends on."""
        raise NotImplementedError  # pragma: no cover

    def validate(self):
        """Validate the project.

        Implementations of this method should raise WorkspaceProjectError if the
        project fails validation.
        """
        raise NotImplementedError  # pragma: no cover

    def sync_command(self, include_dev: bool = True) -> str:
        """Get the command which prepares the project environment.

        This should preferably be a reproducible action (i.e. installing dependencies from a lockfile).
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def new(cls, path: Path):
        """Create a new project at path using default template.

        If an Adapter chooses not to implement this method, the user will be required to
        provide a custom template to initialise a project with this type.
        """
        raise NotImplementedError



import re
from typing import Iterable, List, Optional, Sequence

import click

_BOLD = "\x1b[1m"
_RESET = "\x1b[0m"
_CLEAR = "\x1b[1K\r"


class Colors:
    white = 255
    green = 107
    turquoise = 73
    purple = 140
    yellow = 179
    red = 167


def echo(message=None, file=None, nl=True, err=True, color=None, rewrite=False):
    message = colorize(message)
    if rewrite:
        message = _CLEAR + message
    click.echo(message=message, file=file, nl=nl, err=err, color=color)


def colorize(message: str) -> str:
    return "".join([token for token in _tokenize(message)])


def strip_tags(message: str) -> str:
    return "".join([token for token in _tokenize(message, strip_tags=True)])


def _tokenize(string: str, strip_tags: bool = False):
    style_tags = {
        "header": Colors.turquoise,
        "success": Colors.green,
        "accent": Colors.purple,
        "warning": Colors.yellow,
        "error": Colors.red,
    }
    tokens = {
        "TAGOPEN": r"<[A-Za-z_-]+>",
        "TAGCLOSE": r"</[(A-Za-z_-]+>",
        "DEFAULT": r".",
    }
    token_regex = "|".join(f"(?P<{name}>{regex})" for name, regex in tokens.items())
    stack: List[str] = []

    def parse_tag(tag: str):
        tags = set(style_tags) | {"bold"}
        if tag in tags:
            return tag
        for full_tag in tags:
            if full_tag.startswith(tag):
                return full_tag

    def reset_code():
        style = next((tag for tag in reversed(stack) if tag in style_tags), None)
        bold = "bold" in stack
        result = _RESET
        if style:
            result = result + _color_code(style_tags[style])
        if bold:
            result = result + _BOLD
        return result

    for match in re.finditer(token_regex, string, re.DOTALL):
        kind = match.lastgroup
        token = match.group()
        if kind == "TAGOPEN":
            if strip_tags:
                continue
            tag = parse_tag(token[1:-1])
            if tag == "bold":
                yield _BOLD
            elif tag in style_tags:
                yield _color_code(style_tags[tag])
            else:
                assert False, f"Unknown token {token}"
            stack.append(tag)
        elif kind == "TAGCLOSE":
            if strip_tags:
                continue
            tag = parse_tag(token[2:-1])
            assert stack and (stack[-1] == tag), f"Invalid closing tag {token}"
            stack.pop()
            yield reset_code()
        else:
            yield token
    stack.clear()
    if not strip_tags:
        yield reset_code()


def _color_code(number: int):
    assert number < 256
    return f"\x1b[38;5;{number}m"


def demo_theme():
    echo(
        """<h><b>This is a header</b></h>

This is some regular text, with <a>accented</a> parts.

Let's draw attention to the <s>good</s> things.

<w>This is something to <b>watch out</b> for!</w>

<e>Something here has gone <b>very wrong</b>!</e>

"""
    )  # pragma: no cover


def table(rows: Iterable[Sequence], col_styles: Sequence[Optional[str]] = None) -> str:
    """Convert a list of rows into a justified table."""
    if not rows:
        return ""
    cols = [*zip(*rows)]
    max_lengths = [max([len(strip_tags(str(item))) for item in col]) for col in cols]

    def justify(item, col_idx: int):
        item = str(item)
        if col_styles and col_styles[col_idx]:
            item = f"<{col_styles[col_idx]}>{item}</{col_styles[col_idx]}>"
        tag_length = len(item) - len(strip_tags(item))
        return item.ljust(max_lengths[col_idx] + tag_length)

    def render_row(row: Sequence):
        return "  ".join(justify(item, idx) for idx, item in enumerate(row))

    return "\n".join([render_row(row) for row in rows])



from workspace.cli.theme import strip_tags
from workspace.core.exceptions import WorkspaceBaseError


class WorkspaceCLIError(WorkspaceBaseError):
    def __init__(self, message: str, exit_code: int = 1):
        stripped = strip_tags(message)
        self.args = (stripped,)
        self.display = message
        self.exit_code = 1



import json
from textwrap import indent

import click

from workspace.cli import theme
from workspace.core.models import Workspace


@click.command()
@click.option(
    "--output",
    "-o",
    type=click.Choice(("default", "json")),
    help="Select the output format. By default, just the workspace names will be shown.",
    default="default",
)
def info(output: str = "default"):
    """Display information about the current workspace and its projects."""
    workspace = Workspace.from_path()
    if output == "json":
        theme.echo(
            json.dumps(
                {
                    "path": str(workspace.path),
                    "projects": {
                        name: {"path": str(project.path), "type": project.type}
                        for name, project in workspace.projects.items()
                    },
                    "plugins": workspace.plugins,
                    "template_path": workspace.template_path,
                }
            ),
            err=False,
        )
    else:
        projects_table = theme.table(
            [(name, project.path, f"({project.type})") for name, project in workspace.projects.items()],
            ("bold", "accent", None),
        )
        plugin_list = theme.table([(plugin,) for plugin in workspace.plugins or []], ("bold",))
        template_path = "[" + ", ".join([f"<a>{path}</a>" for path in workspace.template_path or []]) + "]"
        theme.echo(
            f"""
<h>Path</h>: <a>{workspace.path}</a>
<h>Projects</h>:
{indent(projects_table, " " * 4)}
<h>Plugins</h>:
{indent(plugin_list, " " * 4)}
<h>Template Path</h>: {template_path}
""",
            err=False,
        )



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Set, Tuple

import pipenv
from pipfile import Pipfile

from workspace.core.adapter.base import Adapter
from workspace.core.exceptions import WorkspaceBaseError, WorkspaceProjectImproperlyConfigured

assert pipenv  # please the linter - pipenv must be imported to access vendored pipfile.


class PipenvAdapter(Adapter, name="pipenv", command_prefix=("pipenv", "run")):
    @property
    def pipfile_path(self) -> Path:
        return self._project.resolved_path / "Pipfile"

    @property
    def pipfile(self):
        # pipfile is vendored by pipenv, so both imports are necessary.
        return Pipfile.load(self.pipfile_path)

    def dependencies(self, include_dev: bool = True) -> Set[str]:
        pipfile = self.pipfile
        sections = ["default"]
        if include_dev:
            sections.append("develop")
        results = set()
        for section in sections:
            for name, value in pipfile.data[section].items():
                if isinstance(value, dict) and "path" in value:
                    path = (self._project.resolved_path / Path(value["path"])).resolve()
                    project = self._project.root.get_project_by_path(path)
                    if project:
                        results.add(project.name)
        return results

    def validate(self):
        if not (self.pipfile_path.exists() and self.pipfile_path.is_file()):
            raise WorkspaceProjectImproperlyConfigured(f"No Pipfile found in project {self._project.name!r}.")
        try:
            _ = self.pipfile
        except Exception as exc:
            raise WorkspaceProjectImproperlyConfigured(
                f"Error loading pipfile for project {self._project.name!r}: {str(exc)}"
            )

    def sync_command(self, include_dev: bool = True) -> str:
        """Get the command which prepares the project environment."""
        command = "pipenv sync"
        if include_dev:
            command = command + " --dev"
        return command

    def run_args(self, command: str) -> Tuple[str, dict]:
        """Override args for running commands.

        Set PIPENV_IGNORE_VIRTUALENVS to prevent pipenv using the wrong venv in the case that workspace-cli is
        installed with a venv.
        """
        command, kwargs = super().run_args(command)

        env = os.environ.copy()
        env["PIPENV_IGNORE_VIRTUALENVS"] = "1"
        kwargs["env"] = env

        return command, kwargs

    @classmethod
    def new(cls, path: Path):
        try:
            from cookiecutter.main import cookiecutter
        except ImportError:
            raise WorkspaceBaseError("Cookiecutter is not installed - run 'pip install workspace-cli[cookiecutter]'.")
        template_path = Path(__file__).parent.parent.parent / "templates/pipenv"
        try:
            version = sys.version_info
            cookiecutter(
                template=str(template_path),
                output_dir=path.parent,
                extra_context={"project_slug": path.name, "python_version": f"{version.major}.{version.minor}"},
                no_input=True,
            )
        except Exception as exc:
            raise WorkspaceBaseError(
                f"""Failed to initialise pipenv project at {str(path)!r}:
  {str(exc)}
"""
            )



import json
import sys
from typing import Tuple

import click

from workspace.cli import callbacks, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command("list")
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option(
    "--output",
    "-o",
    type=click.Choice(("default", "names", "json")),
    help="Select the output format. By default, just the project names will be shown.",
    default="default",
)
def list_(specifiers: Tuple[str, ...], output: str = "default"):
    """List projects tracked in the workspace."""
    workspace = Workspace.from_path()

    target_set = set(workspace.projects) if click.get_text_stream("stdin").isatty() else set()
    if specifiers:
        target_set = resolve_specifiers(workspace, specifiers)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    for name in sorted(target_set):
        project = workspace.projects[name]
        if output == "json":
            theme.echo(
                json.dumps(
                    {
                        "name": project.name,
                        "type": project.type,
                        "path": str(workspace.path / project.path),
                        "depends_on": list(project.adapter.dependencies()),
                    }
                ),
                err=False,
            )
        elif output == "names":
            theme.echo(project.name, err=False)
        else:
            theme.echo(
                f"""
<h>Name</h>: <b>{project.name}</b>
<h>Type</h>: {project.type}
<h>Path</h>: <a>{workspace.path / project.path}</a>
<h>Dependencies</h>: [{", ".join(project.adapter.dependencies())}]""",
                err=False,
            )



from typing import Any

import click


def consume_stdin(ctx: click.core.Context, param: click.core.Parameter, value: Any):
    """Callback which optionally consumes stdin for arguments."""
    stdin = click.get_text_stream("stdin")
    if not value and not stdin.isatty():
        return param.type_cast_value(ctx, [line.strip() for line in stdin.readlines()])
    else:
        return value



class WorkspaceBaseError(Exception):
    """Base exception for workspaces package.

    Should only be raised directly for unexpected runtime errors.
    """


class WorkspaceError(WorkspaceBaseError):
    """Errors relating to a particular workspace configuration."""


class WorkspaceNotFoundError(WorkspaceError):
    """A workspace file could not be located."""


class WorkspaceValidationError(WorkspaceError):
    """The workspace file could not be parsed."""


class WorkspacePluginError(WorkspaceBaseError):
    """An error relating to an installed plugin."""


class WorkspaceProjectError(WorkspaceBaseError):
    """An error relating to a particular project."""


class WorkspaceProjectImproperlyConfigured(WorkspaceProjectError):
    """A project configuration is invalid."""


class WorkspaceTemplateError(WorkspaceBaseError):
    """An error relating to templates."""



import sys
from pathlib import Path

import click

from workspace.cli import theme
from workspace.core.models import Workspace
from workspace.core.settings import get_settings


@click.command()
@click.option(
    "--path",
    "-p",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, path_type=Path),
    default=None,
)
def init(path: Path = None):
    """Initialise a workspace."""
    path = (path or Path.cwd()).resolve() / get_settings().filename
    if path.exists():
        theme.echo(f"<e>File already exists at <b>{path}</b></e>.")
        sys.exit(1)
    workspace = Workspace(path.parent, projects={})
    workspace.flush()
    theme.echo(f"Created workspace at <a>{path}</a>.")



import re
from fnmatch import fnmatch
from pathlib import Path
from typing import Iterable, Optional, Set

from workspace.core.adapter import get_adapters
from workspace.core.models import Project, Workspace


def resolve_specifiers(workspace: Workspace, specifiers: Iterable[str]) -> Set[str]:
    """Extract project names from given specifiers.

    Specifiers can be exact project names, or glob patterns.
    """
    result = set()
    for specifier in specifiers:
        specifier = specifier.strip()
        for project in workspace.projects:
            if project in result:
                continue
            if specifier == project:
                result.add(project)
                continue
            try:
                if fnmatch(project, specifier):
                    result.add(project)
            except re.error:
                continue
    for name in result:
        workspace.projects[name].adapter.validate()
    return result


def detect_type(workspace: Workspace, path: Path) -> Optional[str]:
    """Detect the type of a project at the given path."""
    for type_name in get_adapters():
        project = Project(
            name="_",
            path=str(Path.cwd().relative_to(workspace.path) / path),
            type=type_name,
            root=workspace,
        )
        try:
            project.adapter.validate()
            return type_name
        except:
            continue
    return None



from __future__ import annotations

import importlib
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

import jsonschema

from workspace.core import exceptions
from workspace.core.adapter import Adapter, get_adapter
from workspace.core.settings import get_settings
from workspace.core.templates import Templates

_PROJECT_CONFIG_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": ["projects"],
    "properties": {
        "projects": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {"path": {"type": "string"}, "type": {"type": "string"}},
            },
        },
        "plugins": {"type": "array", "items": {"type": "string"}},
        "template_path": {"type": "array", "items": {"type": "string"}},
    },
}


@dataclass
class Workspace:
    path: Path
    projects: Dict[str, Project]
    plugins: Optional[List[str]] = None
    template_path: Optional[List[str]] = None

    @classmethod
    def from_path(cls, path: Union[Path, str] = None) -> Workspace:
        path = path or Path.cwd()
        path = Path(path) if isinstance(path, str) else path
        path = path.resolve()
        if path.is_file():
            return cls.load(path)
        for directory in [path, *path.parents]:
            filepath = directory / get_settings().filename
            if filepath.exists() and filepath.is_file():
                return cls.load(filepath)
        raise exceptions.WorkspaceNotFoundError(
            f"No workspace file {get_settings().filename!r} found in '{path}' or its parents."
        )

    @classmethod
    def load(cls, path: Path) -> Workspace:
        text = path.read_text()
        try:
            body = json.loads(text)
        except json.JSONDecodeError:
            raise exceptions.WorkspaceValidationError(f"Workspace file at '{path}' is not valid JSON.")

        try:
            jsonschema.validate(
                schema=_PROJECT_CONFIG_SCHEMA,
                instance=body,
            )
        except jsonschema.ValidationError as exc:
            raise exceptions.WorkspaceValidationError(f"Invalid workspace file at '{path}': {exc.message}")

        kwargs = {"path": path.parent, "projects": {}}
        if "plugins" in body:
            kwargs["plugins"] = body["plugins"]
        if "template_path" in body:
            kwargs["template_path"] = body["template_path"]

        workspace = cls(**kwargs)  # type: ignore[arg-type]
        for name, project in body.get("projects", {}).items():
            workspace.set_project(name, **project)
        workspace._load_plugins()
        return workspace

    @property
    def templates(self) -> Templates:
        return Templates(self)

    def get_project_by_path(self, path: Path) -> Optional[Project]:
        path = path.resolve()
        for project in self.projects.values():
            if project.resolved_path == path:
                return project
        return None

    def set_project(self, name: str, path: str, type: str) -> Project:
        project = Project(name=name, root=self, path=path, type=type)
        self.projects[name] = project
        return project

    def flush(self) -> None:
        output: dict = {
            "projects": {
                project.name: {
                    "path": project.path,
                    "type": project.type,
                }
                for project in self.projects.values()
            }
        }
        if self.plugins is not None:
            output["plugins"] = self.plugins
        if self.template_path is not None:
            output["template_path"] = self.template_path
        with open(self.path / get_settings().filename, "w", encoding="utf-8") as file:
            file.write(json.dumps(output, sort_keys=True, indent=2))

    def _load_plugins(self) -> None:
        for plugin in self.plugins or []:
            try:
                importlib.import_module(plugin)
            except ModuleNotFoundError:
                raise exceptions.WorkspacePluginError(f"Could not find configured plugin {plugin!r}.")


@dataclass
class Project:
    """A particular project within a workspace."""

    name: str
    root: Workspace = field(repr=False)
    path: str
    type: str
    # options (type-specific)

    @property
    def resolved_path(self) -> Path:
        return (self.root.path / Path(self.path)).resolve()

    @property
    def adapter(self) -> Adapter:
        return get_adapter(self.type)(self)



import sys
from pathlib import Path
from typing import Optional, Set, Tuple

import click

from workspace.cli import callbacks, theme
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "paths",
    nargs=-1,
    callback=callbacks.consume_stdin,
    type=click.Path(exists=False, file_okay=True, dir_okay=True, path_type=Path),
)
@click.option(
    "--output",
    "-o",
    type=click.Choice(("lines", "csv")),
    help="Select the output format.",
    default="lines",
)
def reverse(paths: Tuple[Path, ...], output: str = "lines"):
    """Reverse file paths to their respective projects.

    Accepts multiple arguments, and will output the deduplicated set of project names.
    """
    workspace = Workspace.from_path()

    results: Set[str] = set(filter(None, (_reverse_path(workspace, path) for path in paths)))
    if output == "csv":
        theme.echo(",".join(results), err=False)
        sys.exit(0)

    for result in results:
        theme.echo(result, err=False)
    sys.exit(0)


def _reverse_path(workspace: Workspace, path: Path) -> Optional[str]:
    relatives = {}
    path = path.resolve()
    for project in workspace.projects.values():
        try:
            relatives[project.name] = path.relative_to(project.resolved_path)
        except ValueError:
            continue
    if not relatives:
        return None
    return sorted(relatives.items(), key=lambda item: len(item[1].parents))[0][0]



from types import MappingProxyType
from typing import Type

from workspace.core.adapter.base import Adapter
from workspace.core.exceptions import WorkspaceProjectImproperlyConfigured

try:
    from workspace.core.adapter.pipenv import PipenvAdapter

    assert PipenvAdapter
except ImportError:  # pragma: no cover
    pass  # pragma: no cover

try:
    from workspace.core.adapter.poetry import PoetryAdapter

    assert PoetryAdapter
except ImportError:  # pragma: no cover
    pass  # pragma: no cover


__all__ = [
    "Adapter",
    "get_adapter",
]


def get_adapter(name: str) -> Type[Adapter]:
    adapters = get_adapters()
    try:
        return adapters[name]
    except KeyError:
        raise WorkspaceProjectImproperlyConfigured(
            f"No adapter of type {name!r} registered. Available types are {list(adapters)}."
        )


def get_adapters() -> MappingProxyType:
    return MappingProxyType(
        {
            subclass.name: subclass
            for subclass in _all_subclasses(Adapter)
            if hasattr(subclass, "name") and subclass.name
        }
    )


def _all_subclasses(cls: Type):
    """Get all explicit and implicit subclasses of a type."""
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in _all_subclasses(c)])



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import lru_cache


@dataclass
class Settings:
    filename: str = "workspace.json"

    @classmethod
    def from_env(cls):
        env_prefix = "WORKSPACE_"
        params = {
            name: os.environ[env_prefix + name.upper()]
            for name, field in cls.__dataclass_fields__.items()
            if env_prefix + name.upper() in os.environ and field.init
        }
        return cls(**params)


@lru_cache(maxsize=None)
def get_settings() -> Settings:
    return Settings.from_env()



import sys
from typing import Tuple

import click

from workspace.cli import callbacks, runner, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option(
    "--dev/--no-dev",
    "-D/ ",
    type=bool,
    help="Include development dependencies.",
    default=False,
)
@click.option(
    "--parallel/--no-parallel",
    "-p/ ",
    type=bool,
    default=False,
    help="Run the command in parallel.",
)
def sync(specifiers: Tuple[str, ...], dev: bool = False, parallel: bool = False):
    """Sync the environments of the specified projects."""
    workspace = Workspace.from_path()

    if specifiers:
        target_set = resolve_specifiers(workspace, set(specifiers))
    else:
        target_set = set(workspace.projects)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    projects = (workspace.projects[target] for target in target_set)
    project_commands = [(project, project.adapter.sync_command(include_dev=dev)) for project in projects]

    sys.exit(runner.run(project_commands, parallel=parallel))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['workspace',
 'workspace.cli',
 'workspace.cli.commands',
 'workspace.cli.commands.template',
 'workspace.core',
 'workspace.core.adapter',
 'workspace.templates.pipenv.{{ cookiecutter.project_slug }}.{{ '
 'cookiecutter.package_name }}',
 'workspace.templates.poetry.{{ cookiecutter.project_slug }}.{{ '
 'cookiecutter.package_name }}']

package_data = \
{'': ['*'],
 'workspace': ['templates/pipenv/*',
               'templates/pipenv/{{ cookiecutter.project_slug }}/*',
               'templates/poetry/*',
               'templates/poetry/{{ cookiecutter.project_slug }}/*']}

install_requires = \
['click>=8.0,<9.0', 'jsonschema>=4.0.1,<5.0.0']

extras_require = \
{'cookiecutter': ['cookiecutter>=1.7.3,<2.0.0'],
 'pipenv': ['pipenv>=2021.5.29,<2022.0.0'],
 'poetry': ['poetry>=1.1.11,<2.0.0']}

entry_points = \
{'console_scripts': ['workspace = workspace.__main__:run_cli']}

setup_kwargs = {
    'name': 'workspace-cli',
    'version': '0.2.0',
    'description': 'Manage interdependent projects in a workspace.',
    'long_description': '# Workspace\n\n**Workspace** streamlines management and adoption of mono-repositories, by providing a wrapper around multi-repo tooling.\n\nIt was initially implemented to manage Python projects, but can be extended to interpret other types of projects.\n\n## Documentation\n\n- [Home](https://jacksmith15.github.io/workspace-cli/)\n    + [Installation](https://jacksmith15.github.io/workspace-cli/installation/)\n    + [Configuration](https://jacksmith15.github.io/workspace-cli/configuration/)\n    + [Basic usage](https://jacksmith15.github.io/workspace-cli/basic-usage/)\n    + [Templates](https://jacksmith15.github.io/workspace-cli/templates/)\n    + [Plugins](https://jacksmith15.github.io/workspace-cli/plugins/)\n    + [FAQ](https://jacksmith15.github.io/workspace-cli/faq/)\n\n\n## Development\n\nInstall dependencies:\n\n```shell\npyenv shell 3.8.6  # Or other version >= 3.8\npre-commit install  # Configure commit hooks\npoetry install  # Install Python dependencies\n```\n\nRun tests:\n\n```shell\npoetry run inv verify\n```\n\n### Todos\n\n- [ ] Experiment with non-Python workspaces\n- [ ] More detailed plugin documentation\n- [ ] Example workspaces\n- [ ] Test output on shells with limited color support\n- [ ] Labelling projects\n- [ ] Project aliases\n\n## License\n\nThis project is distributed under the MIT license.\n',
    'author': 'Jack Smith',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jacksmith15/workspace-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sys
from pathlib import Path

import click

from workspace.cli import theme
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "action",
    type=click.Choice(["add", "remove"]),
)
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
)
def path(action: str, path: Path):
    workspace = Workspace.from_path()

    path = path.resolve().relative_to(workspace.path)

    if not workspace.template_path:
        workspace.template_path = []

    # TODO: it might be good to print the templates added and removed.
    if action == "add":
        if str(path) in workspace.template_path:
            theme.echo(f"<e>Already configured to detect templates at <b>{path}</b>.</e>")
            sys.exit(1)
        workspace.template_path.append(str(path))
        workspace.flush()
        theme.echo(f"Added <a>{path}</a> to available template directories.")
        sys.exit(0)

    if str(path) not in workspace.template_path:
        theme.echo(f"<e>Not configured to detect templates at <b>{path}</b>.</e>")
        sys.exit(1)

    workspace.template_path.remove(str(path))
    workspace.flush()
    theme.echo(f"Stopped detecting templates at <a>{path}</a>.")



import sys

import click

from workspace.cli import theme
from workspace.cli.commands.add import add
from workspace.cli.commands.dependees import dependees
from workspace.cli.commands.dependencies import dependencies
from workspace.cli.commands.info import info
from workspace.cli.commands.init import init
from workspace.cli.commands.list import list_
from workspace.cli.commands.new import new
from workspace.cli.commands.plugin import plugin
from workspace.cli.commands.remove import remove
from workspace.cli.commands.reverse import reverse
from workspace.cli.commands.run import run
from workspace.cli.commands.sync import sync
from workspace.cli.commands.template import template
from workspace.cli.exceptions import WorkspaceCLIError
from workspace.core.exceptions import WorkspaceBaseError


@click.group()
def cli():
    """Manage interdependent projects in a workspace."""
    pass


for command in (add, dependees, dependencies, info, init, list_, new, plugin, remove, reverse, run, sync, template):
    cli.add_command(command)


def run_cli():
    try:
        cli()
    except WorkspaceCLIError as exc:
        theme.echo(exc.display)
        sys.exit(exc.exit_code)
    except WorkspaceBaseError as exc:
        theme.echo(f"<e>{exc}</e>")
        sys.exit(1)



import click

from workspace.cli import theme
from workspace.core.models import Workspace


@click.command("list")
def list_():
    workspace = Workspace.from_path()
    output = {
        template_path.name: str(template_path.relative_to(workspace.path)) for template_path in workspace.templates
    }
    if not output:
        return
    maxlen = max([len(name) for name in output])
    theme.echo("")
    for template_path in workspace.templates:
        theme.echo(
            f"<b>{template_path.name.ljust(maxlen)}</b>   <a>{template_path.relative_to(workspace.path)}</a>", err=False
        )



import sys
from typing import Tuple

import click

from workspace.cli import callbacks, runner, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option(
    "--command",
    "-c",
    type=str,
    required=True,
    help="The command to execute.",
)
@click.option(
    "--parallel/--no-parallel",
    "-p/ ",
    type=bool,
    default=False,
    help="Run the command in parallel.",
)
def run(specifiers: Tuple[str], command: str, parallel: bool = False):
    """Run a command in each project.

    If no specifiers are provided, the command will be run in all projects.
    """
    workspace = Workspace.from_path()

    target_set = set(workspace.projects) if click.get_text_stream("stdin").isatty() else set()
    if specifiers:
        target_set = resolve_specifiers(workspace, specifiers)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    project_commands = [(workspace.projects[target], command) for target in target_set]
    sys.exit(runner.run(project_commands, parallel=parallel))



import sys
from collections import defaultdict
from functools import lru_cache
from typing import Dict, Set, Tuple

import click

from workspace.cli import callbacks, theme
from workspace.cli.utils import resolve_specifiers
from workspace.core.models import Workspace


@click.command()
@click.argument(
    "specifiers",
    nargs=-1,
    callback=callbacks.consume_stdin,
)
@click.option("--transitive/--no-transitive", type=bool, default=True, help="Only show direct dependees.")
@click.option(
    "--output",
    "-o",
    type=click.Choice(("lines", "csv")),
    help="Select the output format.",
    default="lines",
)
@click.option(
    "--dev/--no-dev",
    "-d/ ",
    type=bool,
    help="Include development dependencies.",
    default=False,
)
def dependees(specifiers: Tuple[str, ...], transitive: bool = True, output: str = "lines", dev: bool = False):
    """Get the set of all projects which depend on the specified projects."""
    workspace = Workspace.from_path()

    target_set = resolve_specifiers(workspace, specifiers)

    if not target_set:
        theme.echo("<w>No projects selected.</w>")
        sys.exit(0)

    # Get a full map of dependees:
    dependee_map = _get_dependee_map(workspace, transitive=transitive, include_dev=dev)

    # Extract the relevant dependees:
    dependees_set = {dependee for target in target_set for dependee in dependee_map[target]}

    # Sort them from most depended on, to least depended on:
    sorted_dependees = [
        name
        for name, _ in sorted(dependee_map.items(), key=lambda item: -len(item[1]))
        if name in dependees_set or name in target_set
    ]

    if output == "csv":
        theme.echo(",".join(sorted_dependees), err=False)
        sys.exit(0)

    for dependee in sorted_dependees:
        theme.echo(dependee, err=False)
    sys.exit(0)


def _get_dependee_map(
    workspace: Workspace,
    transitive: bool = True,
    include_dev: bool = False,
) -> Dict[str, Set[str]]:
    """Get a mapping of projects to their dependees."""

    direct_map = defaultdict(set)
    for name, project in workspace.projects.items():
        for dependency in project.adapter.dependencies(include_dev=include_dev):
            direct_map[dependency].add(name)
    if not transitive:
        return dict(direct_map)

    @lru_cache(maxsize=None)
    def get_dependees(name: str):
        dependees = set(direct_map[name])
        for dependee in set(dependees):
            dependees |= get_dependees(dependee)
        return dependees

    return {name: get_dependees(name) for name in workspace.projects}



import click

from workspace.cli.commands.template.list import list_
from workspace.cli.commands.template.path import path


@click.group()
def template():
    """Commands for managing project templates."""
    pass


template.add_command(list_)
template.add_command(path)



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Optional, Set, Tuple

from poetry.core.factory import Factory
from poetry.core.pyproject.exceptions import PyProjectException
from poetry.core.pyproject.toml import PyProjectTOML

from workspace.core.adapter.base import Adapter
from workspace.core.exceptions import WorkspaceBaseError, WorkspaceProjectImproperlyConfigured


class PoetryAdapter(Adapter, name="poetry", command_prefix=("poetry", "run")):
    @property
    def pyproject_path(self) -> Path:
        return self._project.resolved_path / "pyproject.toml"

    @property
    def pyproject(self) -> PyProjectTOML:
        return PyProjectTOML(path=self.pyproject_path)

    def dependencies(self, include_dev: bool = True) -> Set[str]:
        self.validate()
        pyproject = self.pyproject.file.read()
        dependencies = pyproject["tool"]["poetry"]["dependencies"]

        results = set()
        sections = ["dependencies"]
        if include_dev:
            sections.append("dev-dependencies")
        for section in sections:
            for key, value in dependencies.items():
                if isinstance(value, dict) and "path" in value:
                    path = (self._project.resolved_path / Path(value["path"])).resolve()
                    project = self._project.root.get_project_by_path(path)
                    if project:
                        results.add(project.name)
        return results

    def validate(self):
        if not (self.pyproject_path.exists() and self.pyproject_path.is_file()):
            raise WorkspaceProjectImproperlyConfigured(f"No pyproject.toml found in project {self._project.name!r}.")
        error_message: Optional[str] = None
        try:
            local_config = self.pyproject.poetry_config
        except PyProjectException as exc:
            error_message = "  - {}\n".format(str(exc))
        else:
            check_result = Factory.validate(local_config)
            if check_result["errors"]:
                error_message = ""
                for error in check_result["errors"]:
                    error_message += "  - {}\n".format(error)

        if error_message:
            raise WorkspaceProjectImproperlyConfigured(
                f"The Poetry configuration at {self.pyproject_path} is invalid:\n" + error_message
            )

    def sync_command(self, include_dev: bool = True) -> str:
        """Get the command which prepares the project environment."""
        command = "poetry install"
        if not include_dev:
            command = command + " --no-dev"
        return command

    def run_args(self, command: str) -> Tuple[str, dict]:
        """Get modified command and kwargs that should be used when running inside the project.

        Deactivate any active virtual environments when running commands in the project environment, otherwise poetry
        will use that instead of managing its own.
        """
        command, kwargs = super().run_args(command)

        env = os.environ.copy()
        venv_path = env.get("VIRTUAL_ENV", None)
        if not venv_path:
            return command, kwargs

        del env["VIRTUAL_ENV"]
        env["PATH"] = ":".join([path for path in env.get("PATH", "").split(":") if path != f"{venv_path}/bin"])
        kwargs["env"] = env
        return command, kwargs

    @classmethod
    def new(cls, path: Path):
        try:
            from cookiecutter.main import cookiecutter
        except ImportError:
            raise WorkspaceBaseError("Cookiecutter is not installed - run 'pip install workspace-cli[cookiecutter]'.")
        template_path = Path(__file__).parent.parent.parent / "templates/poetry"
        try:
            version = sys.version_info
            cookiecutter(
                template=str(template_path),
                output_dir=path.parent,
                extra_context={"project_slug": path.name, "python_version": f"{version.major}.{version.minor}"},
                no_input=True,
            )
        except Exception as exc:
            raise WorkspaceBaseError(
                f"""Failed to initialise poetry project at {str(path)!r}:
  {str(exc)}
"""
            )
